import numpy as np
import warnings
import dill
import configparser

from pathlib import Path
from tqdm import tqdm
from typing import cast
from collections import Counter
from scipy.stats import multivariate_normal as mn

from figaro.transform import transform_to_probit
from figaro.exceptions import FIGAROException

#-–––––––––-#
# Utilities #
#-----------#

def recursive_grid(bounds, n_pts, get_1d = False):
    """
    Recursively generates the n-dimensional grid points (extremes are excluded).
    
    Arguments:
        :list-of-lists bounds: extremes for each dimension (excluded)
        :int n_pts:            number of points for each dimension
        :bool get_1d:          return list of 1d-arrays (one per dimension)
        
    Returns:
        :np.ndarray: grid
        :np.ndarray: differential for each grid
        :np.ndarray: list of 1d-arrays (one per dimension)
    """
    bounds = np.atleast_2d(bounds)
    n_pts  = np.atleast_1d(n_pts)
    if len(bounds) == 1:
        d  = np.linspace(bounds[0,0], bounds[0,1], n_pts[0]+2)[1:-1]
        dD = d[1]-d[0]
        if get_1d:
            return np.atleast_2d(d).T, [dD], [d]
        return np.atleast_2d(d).T, [dD]
    else:
        if get_1d:
            grid_nm1, diff, l_1d = recursive_grid(np.array(bounds)[1:], n_pts[1:], get_1d)
        else:
            grid_nm1, diff = recursive_grid(np.array(bounds)[1:], n_pts[1:], get_1d)
        d = np.linspace(bounds[0,0], bounds[0,1], n_pts[0]+2)[1:-1]
        diff.insert(0, d[1]-d[0])
        grid     = []
        for di in d:
            for gi in grid_nm1:
                grid.append([di,*gi])
        if get_1d:
            l_1d.insert(0, d)
            return np.array(grid), diff, l_1d
        return np.array(grid), diff

def rejection_sampler(n_draws, f, bounds, selfunc = None):
    """
    1D rejection sampler, allows for a selection function
    
    Arguments:
        :int n_draws:      number of draws
        :callable f:       probability density to sample from
        :iterable bounds:  upper and lower bound
        :callable selfunc: selection function, must support numpy arrays
    
    Returns:
        :np.ndarray: samples
    """
    n_draws = int(n_draws)
    if selfunc is None:
        selfunc = lambda x: 1
    x   = np.linspace(bounds[0], bounds[1], 1000)
    top = np.max(f(x)*selfunc(x))
    samples = []
    while len(samples) < n_draws:
        pts   = np.random.uniform(bounds[0], bounds[1], size = n_draws)
        probs = f(pts)*selfunc(pts)
        h     = np.random.uniform(0, top, size = n_draws)
        samples.extend(pts[np.where(h < probs)])
    return np.array(samples).flatten()[:n_draws]

def get_priors(bounds, samples = None, mean = None, std = None, cov = None, df = None, k = None, a = None, scale = 5., probit = True, hierarchical = False):
    """
    This method takes the prior parameters for the Normal-Inverse-Wishart distribution in the natural space and returns them as parameters in the probit space, ordered as required by FIGARO. In the following, D will denote the dimensionality of the inferred distribution.

    Four parameters are returned:
    * df, is the number of degrees of freedom for the Inverse Wishart distribution,. It must be greater than D+1. If this parameter is None or does not satisfy the condition df > D+1, the default value D+2 is used;
    * k is the scale parameter for the multivariate Normal distribution. Suggested values are  k <~ 1e-1. If None, the default value 1e-2 is used.
    * mu is the mean of the multivariate Normal distribution. It can be either estimated from the available samples or passed directly as a 1D array with length D (the keyword argument mean overrides the samples). If None, the default value 0 (corresponding to the parameter space center) is used.
    * L is the expected value for the Inverse Wishart distribution. This parameter can be either (in descending priority order):
        * passed as 2D array with shape (D,D), the covariance matrix - keyword cov;
        * passed as 1D array with shape (D,) or double: vector of standard deviations (if double, it assumes that the same std has to be used for all dimensions) - keyword std;
        * estimated from samples - keyword samples.
       
    The order in which they are returned is (k,L,df,mu).
    
    Arguments:
        :np.ndarray bounds:              boundaries for probit transformation
        :np.ndarray samples:             2D array with samples
        :double or np.ndarray mean:      mean [DPGMM]
        :double or np.ndarray std:       expected standard deviation (if double, the same std is used for all dimensions, if np.ndarray must match the number of dimensions) [DPGMM and (H)DPGMM]
        :np.ndarray cov:                 covariance matrix [DPGMM]
        :int df:                         degrees of freedom for Inverse Wishart distribution [DPGMM]
        :double k:                       scale parameter for Normal distribution [DPGMM]
        :double a:                       shape parameter for the Inverse Gamma distribution [(H)DPGMM]
        :double scale:                   fraction of samples std [DPGMM]
        :bool probit:                    whether the probit transformation will be applied or not
        :bool hierarchical:              returns the prior pars for (H)DPGMM rather than for DPGMM
        
    Returns:
        :tuple: prior parameters ordered as in (H)/DPGMM
    """
    bounds = np.atleast_2d(bounds)
    dim = len(bounds)
    if samples is not None:
        if len(np.shape(samples)) < 2:
            samples = np.atleast_2d(samples).T
        if probit:
            probit_samples = transform_to_probit(samples, bounds)
    if hierarchical:
        if std is not None:
            out_sigma = np.atleast_2d(std)*np.ones((1, dim))
            if probit:
                out_sigma = transform_to_probit(np.mean(bounds, axis = -1)+out_sigma, bounds)
        elif samples is not None:
            if probit:
                out_sigma = np.sqrt(np.diag(np.atleast_2d(np.cov(probit_samples.T))))
            else:
                out_sigma = np.sqrt(np.diag(np.atleast_2d(np.cov(samples.T))))
            out_sigma = out_sigma/scale
        else:
            out_sigma = np.diff(bounds, axis = -1)/10.
            if probit:
                out_sigma = transform_to_probit(np.mean(bounds, axis = -1)+out_sigma, bounds)
        out_sigma = out_sigma.flatten()
        if a is not None:
            out_a = a
        else:
            out_a = 2.
            
        return (out_sigma, out_a)
    
    else:
        # DF
        if df is not None and df > dim+2:
            df_out = df
        else:
            df_out = dim+2
            
        draw_flag = False
        # Mu
        if mean is not None:
            mean = np.atleast_1d(mean)
            if not np.prod(bounds[:,0] < mean) & np.prod(mean < bounds[:,1]):
                raise ValueError("Mean is outside of the given bounds")
            if probit:
                mu_out = transform_to_probit(mean, bounds)
            else:
                mu_out = mean
        elif samples is not None:
            if probit:
                mu_out = np.atleast_1d(np.mean(probit_samples, axis = 0))
            else:
                mu_out = np.atleast_1d(np.mean(samples, axis = 0))
        else:
            if probit:
                mu_out = np.zeros(dim)
            else:
                mu_out = np.atleast_1d(np.mean(bounds, axis = 1))
        # L
        if cov is not None:
            L_out = cov
            if probit:
                draw_flag = True
        elif std is not None:
            L_out = np.identity(dim)*std**2
            if probit:
                draw_flag = True
        elif samples is not None:
            if probit:
                L_out = np.atleast_2d(np.cov(probit_samples.T))
            else:
                L_out = np.atleast_2d(np.cov(samples.T))
            L_out = np.identity(dim)*np.diag(L_out/scale**2)
        else:
            if probit:
                L_out = np.identity(dim)*0.2**2
            else:
                L_out = np.identity(dim)*(np.diff(bounds, axis = 1)/10)**2
        if draw_flag:
            ss = mn(np.mean(bounds, axis = -1), L_out).rvs(3000)
            if dim == 1:
                ss = np.atleast_2d(ss).T
            # Keeping only samples within bounds
            ss = ss[np.where((np.prod(bounds[:,0] < ss, axis = 1) & np.prod(ss < bounds[:,1], axis = 1)))]
            probit_samples = transform_to_probit(ss, bounds)
            L_out = np.atleast_2d(np.cov(probit_samples.T))
        # k
        if k is not None:
            k_out = k
        else:
            s, log_k_out = np.linalg.slogdet(L_out)
            k_out = np.exp(log_k_out/dim)
        return (k_out, L_out, df_out, mu_out)

def rvs_median(draws, n_draws):
    """
    Generates samples from median distribution of a set of draws.
    
    Arguments:
        :iterable draws: container for mixture instances
        :int n_draws:    number of samples
    
    Returns:
        :np.ndarray: samples
    """
    idx = np.random.choice(np.arange(len(draws)), size = int(n_draws))
    ctr = Counter(idx)
    samples = np.empty(shape = (1, draws[0].dim))
    for i, n in zip(ctr.keys(), ctr.values()):
        samples = np.concatenate((samples, draws[i].rvs(n)))
    return samples[1:]
    
def make_single_gaussian_mixture(mu, cov, bounds, out_folder = '.', save = False, n_samps = 3000, probit = True):
    """
    Builds mixtures composed of a single Gaussian distribution.
    WARNING: due to the probit coordinate change, a Gaussian distribution in the natural space does not correspond to a Gaussian distribution in the probit space.
    The resulting distributions, therefore, are just an approximation. This approximation holds for distributions which are far from boundaries.
    In general, a more robust (but slower) approach would be to draw samples from each original Gaussian distribution and to use them to make a hierarchical inference.
    
    Arguments:
        :np.ndarray mu:     mean for each Gaussian distribution
        :np.ndarray cov:    covariance matrix for each Gaussian distribution
        :np.ndarray bounds: boundaries for probit transformation
        :str out_folder:    output folder
        :bool save:         whether to save the draws or not
        :int n_samps:       number of samples to estimate mean and covariance in probit space
    
    Returns:
        :np.ndarray: mixtures
    """
    # Here to avoid circular import
    from figaro.mixture import mixture
    bounds = np.atleast_2d(bounds)
    
    out_folder = Path(out_folder)
    if not out_folder.exists():
        out_folder.mkdir()
    
    draws_folder = Path(out_folder, 'draws')
    if not draws_folder.exists():
        draws_folder.mkdir()
    
    events_folder = Path(out_folder, 'events')
    if not events_folder.exists():
        events_folder.mkdir()
    
    if len(cov.shape) == 1:
        cov = np.atleast_2d(cov).T
    
    mixtures = []
    for i, (m, c) in enumerate(zip(mu, cov)):
        if probit:
            ss = np.atleast_2d(mn(m, c).rvs(n_samps))
            # 1D issue
            if c.shape == (1,) or c.shape == (1,1):
                ss = ss.T
            # Keeping only samples within bounds
            ss = ss[np.where((np.prod(bounds[:,0] < ss, axis = 1) & np.prod(ss < bounds[:,1], axis = 1)))]
            if save:
                np.savetxt(Path(events_folder, 'event_{0}.txt'.format(i+1)), ss)
            # Probit samples
            p_ss = transform_to_probit(ss, bounds)
            mm = np.mean(p_ss, axis = 0)
            cc = np.atleast_2d(np.cov(p_ss.T))
        else:
            mm = m
            cc = c
            if save:
                ss = np.atleast_2d(mn(m, c).rvs(n_samps))
                if c.shape == (1,1):
                    ss = ss.T
                np.savetxt(Path(events_folder, 'event_{0}.txt'.format(i+1)), ss)
        mix = mixture(np.atleast_2d([mm]), np.atleast_3d([cc]), np.ones(1), bounds, len(bounds), 1, None, probit = probit)
        mixtures.append([mix])
    
    mixtures = np.array(mixtures)
    
    if save:
        with open(Path(draws_folder, 'posteriors_single_event.pkl'), 'wb') as f:
            dill.dump(mixtures, f)
    
    return mixtures

#-------------#
#   Options   #
#-------------#

def save_options(options, out_folder, name = None):
    """
    Saves options for the run (reproducibility)
    
    Arguments:
        :obj options:            options
        :str or Path out_folder: folder where to save the option file
    """
    if name is None:
        filename = 'options_log.ini'
    else:
        filename = 'options_log_{0}.ini'.format(name)
    with open(Path(out_folder, filename), 'w') as logfile:
        logfile.write('[OPTIONS]\n')
        for key, val in zip(vars(options).keys(), vars(options).values()):
            logfile.write('{0} = {1}\n'.format(key,val))

def load_options(opts, file):
    """
    Loads options for the run (reproducibility)
    
    Arguments:
        :obj opts:         options object
        :str or Path file: file with options
    
    Returns:
        :obj: options
    """
    with open(file, 'r') as logfile:
        config = configparser.RawConfigParser()
        config.read(file)
        opts_dict = dict(config.items('OPTIONS'))
    for key in opts_dict.keys():
        if opts_dict[key] in ['True', 'False']:
            if opts_dict[key] == 'True':
                opts_dict[key] = True
            if opts_dict[key] == 'False':
                opts_dict[key] = False
        if key == 'bounds':
            opts.bounds = str(opts_dict['bounds'])
        else:
            try:
                exec('opts.{0} = opts.{0}.__class__(opts_dict["{0}"])'.format(key))
            except TypeError:
                exec('opts.{0} = opts.{0}.__class__()'.format(key))
    return opts



import numpy
from setuptools import setup, find_packages
from setuptools import Extension
from setuptools.command.build_ext import build_ext as _build_ext
from codecs import open
from os import path
from distutils.extension import Extension
import os
import warnings

lal_flag = True
try:
    import lal
except ModuleNotFoundError:
    lal_flag = False

ray_flag = True
try:
    import ray
except ModuleNotFoundError:
    ray_flag = False

try:
    from Cython.Build import cythonize
except ImportError:
    raise ImportError("Cython not found. Please install it via\n\tpip install Cython")

if os.environ['CONDA_DEFAULT_ENV'] == 'igwn-py39':
    requirements = ['imageio']
else:
    with open("requirements.txt") as requires_file:
        requirements = requires_file.read().split("\n")

with open("README.md") as readme_file:
    long_description = readme_file.read()

# see https://stackoverflow.com/a/21621689/1862861 for why this is here
class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        self.include_dirs.append(numpy.get_include())

ext_modules=[
             Extension("figaro.cumulative",
                       sources=[os.path.join("figaro","cumulative.pyx")],
                       libraries=["m"], # Unix-like specific
                       extra_compile_args=["-O3","-ffast-math"],
                       include_dirs=['figaro', numpy.get_include()]
                       ),
            ]
if lal_flag:
    if "LAL_PREFIX" in os.environ:
        # Older LAL installations requires this
        lal_prefix     = os.environ.get("LAL_PREFIX")
        lal_includes   = lal_prefix+"/include"
        lal_libs       = lal_prefix+"/lib"
        ext_modules.append(Extension("figaro.cosmology",
                          sources=[os.path.join("figaro","cosmology.pyx")],
                          libraries=["m", "lal"], # Unix-like specific
                          library_dirs = [lal_libs],
                          extra_compile_args=["-O3","-ffast-math"],
                          include_dirs=['figaro', lal_includes, numpy.get_include()]
                          ))
    else:
        ext_modules.append(Extension("figaro.cosmology",
                           sources=[os.path.join("figaro","cosmology.pyx")],
                           libraries=["m", "lal"], # Unix-like specific
                           extra_compile_args=["-O3","-ffast-math"],
                           include_dirs=['figaro', numpy.get_include()]
                           ))

ext_modules = cythonize(ext_modules, compiler_directives={'language_level' : "3"})

scripts = ['figaro-density=figaro.pipelines.probability_density:main',
           'figaro-hierarchical=figaro.pipelines.hierarchical_inference:main',
           'figaro-glade=figaro.pipelines.create_glade:main',
           'figaro-pp_plot=figaro.pipelines.ppplot:main',
           'figaro-mockdata=figaro.pipelines.gen_mock_data:main',
           'figaro-entropy=figaro.pipelines.entropy:main',
           ]
pymodules = ['figaro/pipelines/probability_density',
             'figaro/pipelines/hierarchical_inference',
             'figaro/pipelines/create_glade',
             'figaro/pipelines/ppplot',
             'figaro/pipelines/gen_mock_data',
             'figaro/pipelines/entropy',
             ]

par_scripts = ['figaro-par-hierarchical=figaro.pipelines.par_hierarchical_inference:main',
               'figaro-par-density=figaro.pipelines.par_probability_density:main',
               ]
par_modules = ['figaro/pipelines/par_hierarchical_inference',
               'figaro/pipelines/par_probability_density',
              ]

if ray_flag:
    scripts   = scripts + par_scripts
    pymodules = pymodules + par_modules

setup(
    name = 'figaro',
    description = 'FIGARO: Fast Inference for GW Astronomy, Research & Observations',
    author = 'Stefano Rinaldi, Walter Del Pozzo, Daniele Sanfratello',
    author_email = 'stefano.rinaldi@phd.unipi.it, walter.delpozzo@unipi.it, d.sanfratello@studenti.unipi.it',
    url = 'https://github.com/sterinaldi/figaro',
    python_requires = '>=3.7',
    packages = ['figaro'],
    py_modules = pymodules,
    install_requires=requirements,
    include_dirs = ['figaro', numpy.get_include()],
    setup_requires=['numpy', 'cython'],
    package_data={"": ['*.c', '*.pyx', '*.pxd']},
    ext_modules=ext_modules,
    entry_points = {
        'console_scripts': scripts,
        },
    version='1.0.0',
    long_description=long_description,
    long_description_content_type='text/markdown',
    )

setup(
      name = 'figaro/cumulative',
      ext_modules = cythonize(ext_modules, language_level = "3"),
      include_dirs=['figaro', numpy.get_include()]
      )
if lal_flag:
    setup(
          name = 'figaro/cosmology',
          ext_modules = cythonize(ext_modules, language_level = "3"),
          include_dirs=['figaro', numpy.get_include()]
          )


if not lal_flag:
    warnings.warn("\n\nWARNING: No LAL installation found, please install LAL - see https://wiki.ligo.org/Computing/LALSuiteInstall. Some functions - GW posterior samples loading and catalog loading - won't be available and errors might be raised.\n", stacklevel = 2)
if not ray_flag:
    warnings.warn("\n\nWARNING: Ray is not installed: parallelized pipelines won't be available. If you want to use them, please install Ray (pip install ray) and reinstall FIGARO.\n", stacklevel = 2)



import numpy as np

import optparse as op
import dill
import importlib

from pathlib import Path
from tqdm import tqdm

from figaro.mixture import DPGMM, HDPGMM
from figaro.transform import transform_to_probit
from figaro.utils import save_options, load_options, get_priors
from figaro.plot import plot_median_cr, plot_multidim
from figaro.load import load_data, load_single_event, save_density, load_density

import ray
from ray.util import ActorPool

@ray.remote
class worker:
    def __init__(self, bounds,
                       out_folder_plots,
                       out_folder_draws,
                       ext         = 'pkl',
                       se_sigma    = None,
                       hier_sigma  = None,
                       all_samples = None,
                       label       = None,
                       unit        = None,
                       save_se     = True,
                       MC_draws    = 2000,
                       probit      = True,
                       ):
        self.dim                  = bounds.shape[0]
        self.bounds               = bounds
        self.mixture              = DPGMM(self.bounds, probit = probit)
        self.hierarchical_mixture = HDPGMM(self.bounds,
                                           MC_draws   = MC_draws,
                                           probit     = probit,
                                           prior_pars = get_priors(self.bounds,
                                                                   samples      = all_samples,
                                                                   std          = hier_sigma,
                                                                   probit       = probit,
                                                                   hierarchical = True,
                                                                   )
                                            )
        self.out_folder_plots = out_folder_plots
        self.out_folder_draws = out_folder_draws
        self.save_se          = save_se
        self.label            = label
        self.unit             = unit
        self.posteriors       = None
        self.probit           = probit
        self.ext              = ext

    def run_event(self, pars):
        # Unpack data
        samples, name, n_draws = pars
        # Copying (issues with shuffling)
        ev = np.copy(samples)
        ev.setflags(write = True)
        # Actual inference
        prior_pars = get_priors(self.bounds, samples = ev, probit = self.probit, std = self.se_sigma, hierarchical = False)
        self.mixture.initialise(prior_pars = prior_pars)
        draws      = [self.mixture.density_from_samples(ev) for _ in range(n_draws)]
        # Plots
        plt_bounds = np.atleast_2d([ev.min(axis = 0), ev.max(axis = 0)]).T
        if self.save_se:
            if self.dim == 1:
                plot_median_cr(draws,
                               samples    = ev,
                               bounds     = plt_bounds[0],
                               out_folder = self.out_folder_plots,
                               name       = name,
                               label      = self.label,
                               unit       = self.unit,
                               subfolder  = True,
                               )
            else:
                plot_multidim(draws,
                              samples    = ev,
                              bounds     = plt_bounds,
                              out_folder = self.out_folder_plots,
                              name       = name,
                              labels     = self.label,
                              units      = self.unit,
                              subfolder  = True,
                              )
        # Saving
        save_density(draws, folder = self.out_folder_draws, name = 'draws_'+name, ext = self.ext)
        return draws

    def draw_hierarchical(self):
        return self.hierarchical_mixture.density_from_samples(self.posteriors)
    
    def load_posteriors(self, posteriors):
        self.posteriors = np.copy(posteriors)
        self.posteriors.setflags(write = True)
        for i in range(len(self.posteriors)):
            self.posteriors[i].setflags(write = True)
            
def main():

    parser = op.OptionParser()
    # Input/output
    parser.add_option("-i", "--input", type = "string", dest = "samples_folder", help = "Folder with single-event samples files", default = None)
    parser.add_option("-b", "--bounds", type = "string", dest = "bounds", help = "Density bounds. Must be a string formatted as '[[xmin, xmax], [ymin, ymax],...]'. For 1D distributions use '[xmin, xmax]'. Quotation marks are required and scientific notation is accepted", default = None)
    parser.add_option("-o", "--output", type = "string", dest = "output", help = "Output folder. Default: same directory as samples folder", default = None)
    parser.add_option("-j", dest = "json", action = 'store_true', help = "Save mixtures in json file", default = False)
    parser.add_option("--inj_density", type = "string", dest = "inj_density_file", help = "Python module with injected density - please name the method 'density'", default = None)
    parser.add_option("--selfunc", type = "string", dest = "selfunc_file", help = "Python module with selection function - please name the method 'selection_function'", default = None)
    parser.add_option("--parameter", type = "string", dest = "par", help = "GW parameter(s) to be read from files", default = None)
    parser.add_option("--waveform", type = "string", dest = "wf", help = "Waveform to load from samples file. To be used in combination with --parameter. Accepted values: 'combined', 'imr', 'seob'", default = 'combined')
    # Plot
    parser.add_option("--name", type = "string", dest = "h_name", help = "Name to be given to hierarchical inference files. Default: same name as samples folder parent directory", default = None)
    parser.add_option("-p", "--postprocess", dest = "postprocess", action = 'store_true', help = "Postprocessing", default = False)
    parser.add_option("-s", "--save_se", dest = "save_single_event", action = 'store_true', help = "Save single event plots", default = False)
    parser.add_option("--symbol", type = "string", dest = "symbol", help = "LaTeX-style quantity symbol, for plotting purposes", default = None)
    parser.add_option("--unit", type = "string", dest = "unit", help = "LaTeX-style quantity unit, for plotting purposes", default = None)
    parser.add_option("--hier_samples", type = "string", dest = "true_vals", help = "Samples from hierarchical distribution (true single-event values, for simulations only)", default = None)
    # Settings
    parser.add_option("--draws", type = "int", dest = "n_draws", help = "Number of draws for hierarchical distribution", default = 100)
    parser.add_option("--se_draws", type = "int", dest = "n_se_draws", help = "Number of draws for single-event distribution. Default: same as hierarchical distribution", default = None)
    parser.add_option("--n_samples_dsp", type = "int", dest = "n_samples_dsp", help = "Number of samples to analyse (downsampling). Default: all", default = -1)
    parser.add_option("--exclude_points", dest = "exclude_points", action = 'store_true', help = "Exclude points outside bounds from analysis", default = False)
    parser.add_option("--cosmology", type = "string", dest = "cosmology", help = "Cosmological parameters (h, om, ol). Default values from Planck (2021)", default = '0.674,0.315,0.685')
    parser.add_option("-e", "--events", dest = "run_events", action = 'store_false', help = "Skip single-event analysis", default = True)
    parser.add_option("--se_sigma_prior", dest = "se_sigma_prior", type = "string", help = "Expected standard deviation (prior) for single-event inference - single value or n-dim values. If None, it is estimated from samples", default = None)
    parser.add_option("--sigma_prior", dest = "sigma_prior", type = "string", help = "Expected standard deviation (prior) for hierarchical inference - single value or n-dim values. If None, it is estimated from samples", default = None)
    parser.add_option("--n_parallel", dest = "n_parallel", type = "int", help = "Number of parallel threads", default = 4)
    parser.add_option("--mc_draws", dest = "mc_draws", type = "int", help = "Number of draws for assignment MC integral", default = 2000)
    parser.add_option("--snr_threshold", dest = "snr_threshold", type = "float", help = "SNR threshold for simulated GW datasets", default = None)
    parser.add_option("--far_threshold", dest = "far_threshold", type = "float", help = "FAR threshold for simulated GW datasets", default = None)
    parser.add_option("--no_probit", dest = "probit", action = 'store_false', help = "Disable probit transformation", default = True)
    parser.add_option("--config", dest = "config", type = "string", help = "Config file. Warning: command line options are ignored if provided", default = None)

    (options, args) = parser.parse_args()

    # Paths
    if options.samples_folder is not None:
        options.samples_folder = Path(options.samples_folder).resolve()
    elif options.config is not None:
        options.samples_folder = Path('.').resolve()
    else:
        raise Exception("Please provide path to samples.")
    if options.output is not None:
        options.output = Path(options.output).resolve()
        if not options.output.exists():
            options.output.mkdir(parents=True)
    else:
        options.output = options.samples_folder.parent
    if options.config is not None:
        options.config = Path(options.config).resolve()
    output_plots = Path(options.output, 'plots')
    if not output_plots.exists():
        output_plots.mkdir()
    output_draws = Path(options.output, 'draws')
    if not output_draws.exists():
        output_draws.mkdir()
    # Read hierarchical name
    if options.h_name is None:
        options.h_name = options.samples_folder.parent.parts[-1]
    # File extension
    if options.json:
        options.ext = 'json'
    else:
        options.ext = 'pkl'

    if options.config is not None:
        load_options(options, options.config)
    save_options(options, options.output, name = options.h_name)

    # Read bounds
    if options.bounds is not None:
        options.bounds = np.array(np.atleast_2d(eval(options.bounds)), dtype = np.float64)
    elif options.bounds is None and not options.postprocess:
        raise Exception("Please provide bounds for the inference (use -b '[[xmin,xmax],[ymin,ymax],...]')")

    # Read cosmology
    options.h, options.om, options.ol = (float(x) for x in options.cosmology.split(','))
    # Read parameter(s)
    if options.par is not None:
        options.par = options.par.split(',')
    # Read number of single-event draws
    if options.n_se_draws is None:
        options.n_se_draws = options.n_draws
    if options.se_sigma_prior is not None:
        options.se_sigma_prior = np.array([float(s) for s in options.se_sigma_prior.split(',')])
    if options.sigma_prior is not None:
        options.sigma_prior = np.array([float(s) for s in options.sigma_prior.split(',')])

    # If provided, load injected density
    inj_density = None
    if options.inj_density_file is not None:
        inj_file_name = Path(options.inj_density_file).parts[-1].split('.')[0]
        spec = importlib.util.spec_from_file_location(inj_file_name, options.inj_density_file)
        inj_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(inj_module)
        inj_density = inj_module.density
    #If provided, load selecton function
    selfunc = None
    if options.selfunc_file is not None:
        selfunc_file_name = Path(options.selfunc_file).parts[-1].split('.')[0]
        spec = importlib.util.spec_from_file_location(selfunc_file_name, options.selfunc_file)
        selfunc_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(selfunc_module)
        selfunc = selfunc_module.selection_function
    # If provided, load true values
    true_vals = None
    if options.true_vals is not None:
        options.true_vals = Path(options.true_vals).resolve()
        true_vals, true_name = load_single_event(options.true_vals, par = options.par, h = options.h, om = options.om, ol = options.ol, waveform = options.wf)
        if np.shape(true_vals)[-1] == 1:
            true_vals = true_vals.flatten()
    # Load samples
    events, names = load_data(options.samples_folder, par = options.par, n_samples = options.n_samples_dsp, h = options.h, om = options.om, ol = options.ol, waveform = options.wf, snr_threshold = options.snr_threshold, far_threshold = options.far_threshold)
    try:
        dim = np.shape(events[0][0])[-1]
    except IndexError:
        dim = 1
    if options.exclude_points:
        print("Ignoring points outside bounds.")
        for i, ev in enumerate(events):
            events[i] = ev[np.where((np.prod(options.bounds[:,0] < ev, axis = 1) & np.prod(ev < options.bounds[:,1], axis = 1)))]
        all_samples = np.atleast_2d(np.concatenate(events))
    else:
        # Check if all samples are within bounds
        all_samples = np.atleast_2d(np.concatenate(events))
        if options.probit:
            if not np.alltrue([(all_samples[:,i] > options.bounds[i,0]).all() and (all_samples[:,i] < options.bounds[i,1]).all() for i in range(dim)]):
                raise ValueError("One or more samples are outside the given bounds.")

    # Plot labels
    if dim > 1:
        if options.symbol is not None:
            symbols = options.symbol.split(',')
        else:
            symbols = options.symbol
        if options.unit is not None:
            units = options.unit.split(',')
        else:
            units = options.unit
    else:
        symbols = options.symbol
        units   = options.unit
    
    # Reconstruction
    if not options.postprocess:
        ray.init(num_cpus = options.n_parallel)
        pool = ActorPool([worker.remote(bounds           = options.bounds,
                                        out_folder_plots = output_plots,
                                        out_folder_draws = output_draws,
                                        ext              = options.ext,
                                        se_sigma         = options.se_sigma_prior,
                                        hier_sigma       = options.sigma_prior,
                                        all_samples      = all_samples,
                                        label            = symbols,
                                        unit             = units,
                                        save_se          = options.save_single_event,
                                        MC_draws         = options.mc_draws,
                                        probit           = options.probit,
                                        )
                          for _ in range(options.n_parallel)])
        
        if options.run_events:
            # Run each single-event analysis
            posteriors = []
            for s in tqdm(pool.map_unordered(lambda a, v: a.run_event.remote(v), [[ev, name, options.n_se_draws] for ev, name in zip(events, names)]), total = len(events), desc = 'Events'):
                posteriors.append(s)
            # Save all single-event draws together
            posteriors = np.array(posteriors)
            save_density(posteriors, folder = output_draws, name = 'posteriors_single_event', ext = options.ext)
        else:
            posteriors = load_density(Path(output_draws, 'posteriors_single_event.'+options.ext))
        # Load posteriors
        for s in pool.map(lambda a, v: a.load_posteriors.remote(v), [posteriors for _ in range(options.n_parallel)]):
            pass
        # Run hierarchical analysis
        draws = []
        for s in tqdm(pool.map_unordered(lambda a, v: a.draw_hierarchical.remote(), [_ for _ in range(options.n_draws)]), total = options.n_draws, desc = 'Sampling'):
            draws.append(s)
        draws = np.array(draws)
        # Save draws
        save_density(draws, folder = output_draws, name = 'draws_'+options.h_name, ext = options.ext)
    else:
        draws = load_density(Path(output_draws, 'draws_'+options.h_name+'.'+options.ext))
    # Plot
    if dim == 1:
        plot_median_cr(draws,
                       injected     = inj_density,
                       selfunc      = selfunc,
                       samples      = true_vals,
                       out_folder   = output_plots,
                       name         = options.h_name,
                       label        = options.symbol,
                       unit         = options.unit,
                       hierarchical = True,
                       )
    else:
        plot_multidim(draws,
                      samples      = true_vals,
                      out_folder   = output_plots,
                      name         = options.h_name,
                      labels       = symbols,
                      units        = units,
                      hierarchical = True,
                      )

if __name__ == '__main__':
    main()
